//
//  Copyright (c) 2020
//  Joern Ungermann
//
// based largely on existing kernels in uBLAS (axpy_prod/sparse_prod)
//

#include <iostream>
#include <iomanip>

#include <boost/numeric/ublas/vector.hpp>
#include <boost/numeric/ublas/vector_proxy.hpp>
#include <boost/numeric/ublas/matrix.hpp>
#include <boost/numeric/ublas/matrix_sparse.hpp>
#include <boost/numeric/ublas/matrix_proxy.hpp>
#include <boost/numeric/ublas/triangular.hpp>
#include <boost/numeric/ublas/io.hpp>
#include <boost/numeric/ublas/lu.hpp>

#include <boost/numeric/ublas/traits.hpp>
#include <boost/numeric/ublas/vector.hpp>
#include <boost/numeric/ublas/matrix.hpp>
#include <boost/numeric/ublas/matrix_sparse.hpp>
#include <boost/numeric/ublas/triangular.hpp>
#include <boost/numeric/ublas/io.hpp>
#include <boost/numeric/ublas/operation_sparse.hpp>
#include <boost/numeric/ublas/operation_blocked.hpp>


#include <sys/times.h>
#include <sys/time.h>
#include <sys/stat.h>

#include "utils.hpp"

namespace ublas  = boost::numeric::ublas;

static const double TOL(1.0e-5); ///< Used for comparing two real numbers.
static const int n(2000);           ///< defines the test matrix size
static const int m(2000);           ///< defines the test matrix size
static const int o(2000);           ///< defines the test matrix size

namespace boost {
  namespace numeric {
    namespace ublas{

      // these commented-out kernels did not perform very well

//       template<class M, class E1, class E2>
//       BOOST_UBLAS_INLINE
//       M&
//       axpy_prod2 (const matrix_expression<E1> &e1,
//                   unknown_storage_tag, row_major_tag,
//                   const matrix_expression<E2> &e2,
//                   unknown_storage_tag, row_major_tag,
//                   M &m,
//                   unknown_storage_tag, row_major_tag) {
//         typedef typename M::size_type size_type;
//         typedef typename E1::difference_type difference_type;
//         typedef typename M::value_type value_type;

//         typename E1::const_iterator1       e1it1     (e1 ().begin1());
//         const typename E1::const_iterator1 e1it1_end (e1 ().end1());
//         typename E1::const_iterator2       e1it2;
//         typename E1::const_iterator2       e1it2_end;
//         typename E2::const_iterator2       e2it2;
//         typename E2::const_iterator2       e2it2_end;

//         mapped_vector<value_type> temporary(e2 ().size2 ());
//         typename mapped_vector<value_type>::const_iterator itv;
//         typename mapped_vector<value_type>::const_iterator itv_end;

//         while (e1it1 != e1it1_end) {
//           e1it2     = e1it1.begin();
//           e1it2_end = e1it1.end();
//           const size_type i = e1it2.index1();

//           while (e1it2 != e1it2_end) {
//             const size_type k = e1it2.index2();
//             e2it2     = e2 ().find2 (1, k, 0);
//             e2it2_end = e2 ().find2 (1, k, e1 ().size2 ());
//             while (e2it2 != e2it2_end) {
//               temporary(e2it2.index2()) += (*e1it2) * (*e2it2);
//               ++ e2it2;
//             }
//             ++ e1it2;
//           }

//           itv = temporary.begin();
//           itv_end = temporary.end();
//           while (itv != itv_end) {
//             if (*itv != value_type()) {
//               m(i, itv.index()) = *itv;
//             }
//             ++ itv;
//           }
//           temporary.clear();

//           ++ e1it1;
//         }

//         return m;
//       }

//       template<class M, class E1, class E2>
//       BOOST_UBLAS_INLINE
//       M&
//       axpy_prod2 (const matrix_expression<E1> &e1,
//                   unknown_storage_tag, column_major_tag,
//                   const matrix_expression<E2> &e2,
//                   unknown_storage_tag, column_major_tag,
//                   M &m,
//                   unknown_storage_tag, column_major_tag) {
//         std::cerr << "ccc\n";
//         typedef typename M::size_type size_type;
//         typedef typename E1::difference_type difference_type;
//         typedef typename M::value_type value_type;

//         typedef mapped_vector<value_type> vector_type;

//         typename E2::const_iterator2 e2it2    ( e2 ().begin2());
//         const typename E2::const_iterator2 e2it2_end( e2 ().end2());
//         typename E2::const_iterator1 e2it1;
//         typename E2::const_iterator1 e2it1_end;
//         typename E1::const_iterator1 e1it1;
//         typename E1::const_iterator1 e1it1_end;

//         vector_type temporary(e2 ().size2 ());
//         typename vector_type::const_iterator itv;
//         typename vector_type::const_iterator itv_end;

//         while (e2it2 != e2it2_end) {
//           e2it1     = e2it2.begin();
//           e2it1_end = e2it2.end();
//           while (e2it1 != e2it1_end) {
//             e1it1     = e1 ().find1 (1, 0, e2it1.index1());
//             e1it1_end = e1 ().find1 (1, e2 ().size1 (), e2it1.index1());
//             while (e1it1 != e1it1_end) {
//               temporary(e1it1.index1()) += (*e1it1) * (*e2it1);
//               ++ e1it1;
//             }
//             ++ e2it1;
//           }
//           itv     = temporary.begin();
//           itv_end = temporary.end();
//           //std::cerr << temporary << "\n";
//           while ( itv != itv_end ) {
//             m(itv.index(), e2it2.index2()) = *itv;
//             ++ itv;
//           }
//           temporary.clear();

//           ++ e2it2;
//         }

//         return m;
//       }


//       template<class M, class E1, class E2>
//       BOOST_UBLAS_INLINE
//       M&
//       axpy_prod2 (const matrix_expression<E1> &e1,
//                   unknown_storage_tag, row_major_tag,
//                   const matrix_expression<E2> &e2,
//                   unknown_storage_tag, column_major_tag,
//                   M &m,
//                   dense_proxy_tag, column_major_tag) {
//         typedef typename M::size_type size_type;
//         typedef typename E1::difference_type difference_type;
//         typedef typename M::value_type value_type;

//         typename E1::const_iterator1 e1it1    (e1 ().begin1());
//         typename E1::const_iterator1 e1it1_end(e1 ().end1());
//         typename E1::const_iterator2 e1it2;
//         typename E1::const_iterator2 e1it2_begin;
//         typename E1::const_iterator2 e1it2_end;

//         typename E2::const_iterator1 e2it1;
//         typename E2::const_iterator1 e2it1_end;
//         typename E2::const_iterator2 e2it2;
//         typename E2::const_iterator2 e2it2_end;

//         while (e1it1!=e1it1_end) {
//           e1it2_begin = e1it1.begin();
//           e1it2_end   = e1it1.end();
//           e2it2       = e2 ().begin2();
//           e2it2_end   = e2 ().end2();

//           while (e2it2 != e2it2_end) {
//             e1it2     = e1it2_begin;
//             e2it1     = e2it2.begin();
//             e2it1_end = e2it2.end();

//             double temp = 0;
//             while (e1it2 != e1it2_end && e2it1 != e2it1_end) {
//               const size_type k1 = e1it2.index2();
//               const size_type k2 = e2it1.index1();
//               if (k1 == k2) {
//                 temp += (*e1it2) * (*e2it1);
//                 ++ e1it2;
//                 ++ e2it1;
//               } else if (k1 < k2) {
//                 ++ e1it2;
//               } else {
//                 ++ e2it1;
//               }
//             }
//             m(e1it2.index1(), e2it1.index2()) = temp;
//             ++ e2it2;
//           }
//           ++ e1it1;
//         }

//         return m;
//       }

      template<class M, class E1, class E2>
      BOOST_UBLAS_INLINE
      M&
      axpy_prod2 (const matrix_expression<E1> &e1,
                  unknown_storage_tag, row_major_tag,
                  const matrix_expression<E2> &e2,
                  unknown_storage_tag, row_major_tag,
                  M &m,
                  unknown_storage_tag, row_major_tag) {
        std::cerr <<"ururur";
        typedef typename M::size_type  size_type;
        typedef typename M::value_type value_type;

        typename E1::const_iterator1       e1it1     ( e1 ().begin1() );
        const typename E1::const_iterator1 e1it1_end ( e1 ().end1()   );
        typename E2::const_iterator2       e2it2;
        typename E2::const_iterator2       e2it2_end;

        while (e1it1 != e1it1_end) {
          typename E1::const_iterator2 e1it2     ( e1it1.begin() );
          typename E1::const_iterator2 e1it2_end ( e1it1.end()   );

          const size_type i (e1it2.index1());

          while (e1it2 != e1it2_end) {
            const size_type k (e1it2.index2());
            e2it2     = e2 ().find2 (1, k, 0);
            e2it2_end = e2 ().find2 (1, k, e2 ().size2 ());
            while (e2it2 != e2it2_end) {
              const value_type temp = (*e1it2) * (*e2it2);
              if (temp != value_type())
                m(i, e2it2.index2()) += temp;
              ++ e2it2;
            }
            ++ e1it2;
          }
          ++ e1it1;
        }

        return m;
      }

      template<class M, class E1, class E2>
      BOOST_UBLAS_INLINE
      M&
      axpy_prod2 (const matrix_expression<E1> &e1,
                  unknown_storage_tag, column_major_tag,
                  const matrix_expression<E2> &e2,
                  unknown_storage_tag, column_major_tag,
                  M &m,
                  unknown_storage_tag, column_major_tag) {
        std::cerr <<"ucucuc";
        typedef typename M::size_type size_type;
        typedef typename M::value_type value_type;

        typename E2::const_iterator2 e2it2           ( e2 ().begin2() );
        const typename E2::const_iterator2 e2it2_end ( e2 ().end2()   );
        typename E2::const_iterator1 e2it1;
        typename E2::const_iterator1 e2it1_end;

        while (e2it2 != e2it2_end) {
          e2it1     = e2it2.begin();
          e2it1_end = e2it2.end();
          while (e2it1 != e2it1_end) {
            typename E1::const_iterator1 e1it1     ( e1 ().find1 (1, 0, e2it1.index1()) );
            typename E1::const_iterator1 e1it1_end ( e1 ().find1 (1, e1 ().size1 (), e2it1.index1()) );
            while (e1it1 != e1it1_end) {
              const value_type temp = (*e1it1) * (*e2it1);
              if (temp != value_type())
                m(e1it1.index1(), e2it1.index2()) += temp;
              ++ e1it1;
            }
            ++ e2it1;
          }
          ++ e2it2;
        }

        return m;
      }


      template<class M, class E1, class E2>
      BOOST_UBLAS_INLINE
      M&
      axpy_prod2 (const matrix_expression<E1> &e1,
                  unknown_storage_tag, column_major_tag,
                  const matrix_expression<E2> &e2,
                  unknown_storage_tag, row_major_tag,
                  M &m,
                  dense_proxy_tag,     row_major_tag) {
        std::cerr <<"ucurdr";
        typedef typename M::size_type size_type;
        typedef typename M::value_type value_type;

        typename E1::const_iterator2 e1it2           ( e1 ().begin2() );
        const typename E1::const_iterator2 e1it2_end ( e1 ().end2()   );
        typename E2::const_iterator2 e2it2;
        typename E2::const_iterator2 e2it2_end;

        while (e1it2 != e1it2_end) {
          typename E1::const_iterator1 e1it1     ( e1it2.begin() );
          typename E1::const_iterator1 e1it1_end ( e1it2.end()   );

          while (e1it1 != e1it1_end) {
            e2it2     = e2 ().find2 (1, e1it1.index2(), 0);
            e2it2_end = e2 ().find2 (1, e1it1.index2(), e2 ().size2 ());
            const size_type i ( e1it1.index1() );

            while (e2it2 != e2it2_end) {
              const value_type temp = (*e1it1) * (*e2it2);
              if (temp != value_type())
                m(i, e2it2.index2()) += temp;
              ++ e2it2;
            }
            ++ e1it1;
          }
          ++ e1it2;
        }

        return m;
      }

      template<class M, class E1, class E2>
      BOOST_UBLAS_INLINE
      M&
      axpy_prod2 (const matrix_expression<E1> &e1,
                  unknown_storage_tag, row_major_tag,
                  const matrix_expression<E2> &e2,
                  dense_proxy_tag,     column_major_tag,
                  M &m,
                  unknown_storage_tag, column_major_tag) {
        std::cerr <<"urdcuc";
        typedef typename M::size_type  size_type;
        typedef typename M::value_type value_type;

        size_type size2 (e2 ().size2 ());
        for (size_type j = 0; j < size2; ++ j) {
          typename E1::const_iterator1 e1it1     ( e1 ().begin1() );
          typename E1::const_iterator1 e1it1_end ( e1 ().end1() );
          while (e1it1 != e1it1_end) {
            typename E1::const_iterator2 e1it2     ( e1it1.begin() );
            typename E1::const_iterator2 e1it2_end ( e1it1.end()   );
            value_type temp;
            while (e1it2 != e1it2_end) {
              temp += (*e1it2) * e2 ()(e1it2.index2(), j);
              ++ e1it2;
            }
            if (temp != value_type())
              m(e1it1.index1(), j) = temp;
            ++ e1it1;
          }
        }

        return m;
      }

      template<class M, class E1, class E2>
      BOOST_UBLAS_INLINE
      M&
      axpy_prod2 (const matrix_expression<E1> &e1,
                  dense_proxy_tag,     row_major_tag,
                  const matrix_expression<E2> &e2,
                  unknown_storage_tag, column_major_tag,
                  M &m,
                  unknown_storage_tag, row_major_tag) {
        std::cerr <<"drucur";
        typedef typename M::size_type  size_type;
        typedef typename M::value_type value_type;

        typename E2::const_iterator1 e2it1;
        typename E2::const_iterator1 e2it1_end;
        typename E2::const_iterator2 e2it2;
        typename E2::const_iterator2 e2it2_end;

        size_type size1 (e1 ().size1 ());
        for (size_type i = 0; i < size1; ++ i) {
          e2it2     = e2 ().begin2();
          e2it2_end = e2 ().end2();
          while (e2it2 != e2it2_end) {
            e2it1     = e2it2.begin();
            e2it1_end = e2it2.end();
            value_type temp;
            while (e2it1 != e2it1_end) {
              temp +=  e1 () (i, e2it1.index1()) * (*e2it1);
              ++ e2it1;
            }
            if (temp != value_type())
              m(i, e2it2.index2()) = temp;
            ++ e2it2;
          }
        }

        return m;
      }

      template<class M, class E1, class E2>
      BOOST_UBLAS_INLINE
      M&
      axpy_prod2 (const matrix_expression<E1> &e1,
                  dense_proxy_tag,     row_major_tag,
                  const matrix_expression<E2> &e2,
                  dense_proxy_tag,     column_major_tag,
                  M &m,
                  unknown_storage_tag, column_major_tag) {
        std::cerr <<"d-d-uc";
        typedef typename M::size_type  size_type;
        typedef typename M::value_type value_type;

        size_type size1 (e1 ().size1 ());
        size_type size2 (e2 ().size2 ());
        size_type size3 (e1 ().size2 ());
        for (size_type j = 0; j < size2; ++ j) {
          for (size_type i = 0; i < size1; ++ i) {
            value_type temp;
            for (size_type k = 0; k < size3; ++ k) {
              temp += e1 ()(i, k) * e2 ()(k, j);
            }
            if (temp != value_type())
              m(i, j) = temp;
          }
        }
        return m;
      }

      template<class M, class E1, class E2>
      BOOST_UBLAS_INLINE
      M&
      axpy_prod2 (const matrix_expression<E1> &e1,
                  dense_proxy_tag,     row_major_tag,
                  const matrix_expression<E2> &e2,
                  dense_proxy_tag,     column_major_tag,
                  M &m,
                  unknown_storage_tag, row_major_tag) {
        std::cerr <<"d-d-ur";
        typedef typename M::size_type  size_type;
        typedef typename M::value_type value_type;

        const size_type size1 (e1 ().size1 ());
        const size_type size2 (e2 ().size2 ());
        const size_type size3 (e1 ().size2 ());

        for (size_type i = 0; i < size1; ++ i) {
          for (size_type j = 0; j < size2; ++ j) {
            value_type temp;
            for (size_type k = 0; k < size3; ++ k) {
              temp += e1 ()(i, k) * e2 ()(k, j);
            }
            if (temp != value_type())
              m(i, j) = temp;
          }
        }
        return m;
      }

      // proto
      template<class M, class E1, class E2>
      BOOST_UBLAS_INLINE
      M&
      axpy_prod2 (const matrix_expression<E1> &e1,
                  dense_proxy_tag, row_major_tag /* dont't care */,
                  const matrix_expression<E2> &e2,
                  dense_proxy_tag, row_major_tag,
                  M &m,
                  dense_proxy_tag, row_major_tag) {
        std::cerr <<"drdrdr";
        typedef typename M::size_type  size_type;
        typedef typename M::value_type value_type;

        const size_type size1 (e1 ().size1 ());
        const size_type size2 (e1 ().size2 ());

        for (size_type i = 0; i < size1; ++ i) {
          for (size_type j = 0; j < size2; ++ j) {
            row (m, i).plus_assign (e1 () (i, j) * row (e2 (), j));
          }
        }
        return m;
      }

      // proto
      template<class M, class E1, class E2>
      BOOST_UBLAS_INLINE
      M&
      axpy_prod2 (const matrix_expression<E1> &e1,
                  dense_proxy_tag, column_major_tag,
                  const matrix_expression<E2> &e2,
                  dense_proxy_tag, column_major_tag /*don't care*/,
                  M &m,
                  dense_proxy_tag, column_major_tag) {
        std::cerr <<"dcdcdc";
        typedef typename M::size_type  size_type;
        typedef typename M::value_type value_type;

        const size_type size1 (e2 ().size1 ());
        const size_type size2 (e2 ().size2 ());

        for (size_type j = 0; j < size2; ++ j) {
          for (size_type i = 0; i < size1; ++ i) {
            column (m, j).plus_assign (e2 () (i, j) * column (e1 (), i));
          }
        }
        return m;
      }

      template<class M, class E1, class E2>
      BOOST_UBLAS_INLINE
      M&
      axpy_prod2 (const matrix_expression<E1> &e1,
                  dense_proxy_tag, row_major_tag,
                  const matrix_expression<E2> &e2,
                  dense_proxy_tag, row_major_tag,
                  M &m,
                  dense_proxy_tag, column_major_tag) {
        return axpy_prod2(e1, dense_proxy_tag(),     row_major_tag(),
                          e2, dense_proxy_tag(),     column_major_tag(),
                          m,  unknown_storage_tag(), column_major_tag() );
      }

      template<class M, class E1, class E2>
      BOOST_UBLAS_INLINE
      M&
      axpy_prod2 (const matrix_expression<E1> &e1,
                  dense_proxy_tag,     row_major_tag,
                  const matrix_expression<E2> &e2,
                  dense_proxy_tag,     row_major_tag,
                  M &m,
                  unknown_storage_tag, row_major_tag) {
        return axpy_prod2(e1, dense_proxy_tag(),     row_major_tag(),
                          e2, dense_proxy_tag(),     column_major_tag(),
                          m,  unknown_storage_tag(), row_major_tag() );
      }

      template<class M, class E1, class E2>
      BOOST_UBLAS_INLINE
      M&
      axpy_prod2 (const matrix_expression<E1> &e1,
                  dense_proxy_tag,     row_major_tag,
                  const matrix_expression<E2> &e2,
                  dense_proxy_tag,     row_major_tag,
                  M &m,
                  unknown_storage_tag, column_major_tag) {
        return axpy_prod2(e1, dense_proxy_tag(),     row_major_tag(),
                          e2, dense_proxy_tag(),     column_major_tag(),
                          m,  unknown_storage_tag(), column_major_tag() );
      }

      template<class M, class E1, class E2>
      BOOST_UBLAS_INLINE
      M&
      axpy_prod2 (const matrix_expression<E1> &e1,
                  dense_proxy_tag, row_major_tag,
                  const matrix_expression<E2> &e2,
                  dense_proxy_tag, column_major_tag,
                  M &m,
                  dense_proxy_tag, row_major_tag) {
        return axpy_prod2(e1, dense_proxy_tag(),     row_major_tag(),
                          e2, dense_proxy_tag(),     column_major_tag(),
                          m,  unknown_storage_tag(), row_major_tag() );
      }

      template<class M, class E1, class E2>
      BOOST_UBLAS_INLINE
      M&
      axpy_prod2 (const matrix_expression<E1> &e1,
                  dense_proxy_tag,     row_major_tag,
                  const matrix_expression<E2> &e2,
                  unknown_storage_tag, column_major_tag,
                  M &m,
                  dense_proxy_tag,     row_major_tag) {
        return axpy_prod2(e1, unknown_storage_tag(), column_major_tag(),
                          e2, unknown_storage_tag(), column_major_tag(),
                          m,  unknown_storage_tag(), column_major_tag());
      }

      template<class M, class E1, class E2>
      BOOST_UBLAS_INLINE
      M&
      axpy_prod2 (const matrix_expression<E1> &e1,
                  dense_proxy_tag, column_major_tag,
                  const matrix_expression<E2> &e2,
                  dense_proxy_tag, row_major_tag,
                  M &m,
                  dense_proxy_tag, row_major_tag) {
        return axpy_prod2(e1, dense_proxy_tag(), row_major_tag(),
                          e2, dense_proxy_tag(), row_major_tag(),
                          m,  dense_proxy_tag(), row_major_tag() );
      }

      template<class M, class E1, class E2>
      BOOST_UBLAS_INLINE
      M&
      axpy_prod2 (const matrix_expression<E1> &e1,
                  dense_proxy_tag, column_major_tag,
                  const matrix_expression<E2> &e2,
                  dense_proxy_tag, row_major_tag,
                  M &m,
                  dense_proxy_tag, column_major_tag) {
        return axpy_prod2(e1, dense_proxy_tag(), column_major_tag(),
                          e2, dense_proxy_tag(), column_major_tag(),
                          m,  dense_proxy_tag(), column_major_tag() );
      }

      template<class M, class E1, class E2>
      BOOST_UBLAS_INLINE
      M&
      axpy_prod2 (const matrix_expression<E1> &e1,
                  dense_proxy_tag,     column_major_tag,
                  const matrix_expression<E2> &e2,
                  dense_proxy_tag,     row_major_tag,
                  M &m,
                  unknown_storage_tag, column_major_tag) {
        return axpy_prod2(e1, dense_proxy_tag(),     row_major_tag(),
                          e2, dense_proxy_tag(),     column_major_tag(),
                          m,  unknown_storage_tag(), column_major_tag() );
      }

      template<class M, class E1, class E2>
      BOOST_UBLAS_INLINE
      M&
      axpy_prod2 (const matrix_expression<E1> &e1,
                  dense_proxy_tag, column_major_tag,
                  const matrix_expression<E2> &e2,
                  dense_proxy_tag, column_major_tag,
                  M &m,
                  dense_proxy_tag, row_major_tag) {
        return axpy_prod2(e1, dense_proxy_tag(),     row_major_tag(),
                          e2, dense_proxy_tag(),     column_major_tag(),
                          m,  unknown_storage_tag(), row_major_tag() );
      }

      template<class M, class E1, class E2>
      BOOST_UBLAS_INLINE
      M&
      axpy_prod2 (const matrix_expression<E1> &e1,
                  dense_proxy_tag,     column_major_tag,
                  const matrix_expression<E2> &e2,
                  dense_proxy_tag,     column_major_tag,
                  M &m,
                  unknown_storage_tag, row_major_tag) {
        return axpy_prod2(e1, dense_proxy_tag(),     row_major_tag(),
                          e2, dense_proxy_tag(),     column_major_tag(),
                          m,  unknown_storage_tag(), row_major_tag() );
      }

      template<class M, class E1, class E2>
      BOOST_UBLAS_INLINE
      M&
      axpy_prod2 (const matrix_expression<E1> &e1,
                  dense_proxy_tag,     column_major_tag,
                  const matrix_expression<E2> &e2,
                  dense_proxy_tag,     column_major_tag,
                  M &m,
                  unknown_storage_tag, column_major_tag) {
        return axpy_prod2(e1, dense_proxy_tag(),     row_major_tag(),
                          e2, dense_proxy_tag(),     column_major_tag(),
                          m,  unknown_storage_tag(), column_major_tag() );
      }

      template<class M, class E1, class E2>
      BOOST_UBLAS_INLINE
      M&
      axpy_prod2 (const matrix_expression<E1> &e1,
                  dense_proxy_tag,     column_major_tag,
                  const matrix_expression<E2> &e2,
                  dense_proxy_tag,     row_major_tag,
                  M &m,
                  unknown_storage_tag, row_major_tag) {
        return axpy_prod2(e1, dense_proxy_tag(),     row_major_tag(),
                          e2, dense_proxy_tag(),     column_major_tag(),
                          m,  unknown_storage_tag(), row_major_tag() );
      }

      template<class M, class E1, class E2>
      BOOST_UBLAS_INLINE
      M&
      axpy_prod2 (const matrix_expression<E1> &e1,
                  dense_proxy_tag,     column_major_tag,
                  const matrix_expression<E2> &e2,
                  unknown_storage_tag, column_major_tag,
                  M &m,
                  dense_proxy_tag,     row_major_tag) {
        return axpy_prod2(e1, dense_proxy_tag(),     row_major_tag(),
                          e2, unknown_storage_tag(), column_major_tag(),
                          m,  unknown_storage_tag(), row_major_tag() );
      }

      template<class M, class E1, class E2>
      BOOST_UBLAS_INLINE
      M&
      axpy_prod2 (const matrix_expression<E1> &e1,
                  dense_proxy_tag,     column_major_tag,
                  const matrix_expression<E2> &e2,
                  unknown_storage_tag, column_major_tag,
                  M &m,
                  unknown_storage_tag, row_major_tag) {
        return axpy_prod2(e1, dense_proxy_tag(),     row_major_tag(),
                          e2, unknown_storage_tag(), column_major_tag(),
                          m,  unknown_storage_tag(), row_major_tag() );
      }

      template<class M, class E1, class E2>
      BOOST_UBLAS_INLINE
      M&
      axpy_prod2 (const matrix_expression<E1> &e1,
                  unknown_storage_tag, row_major_tag,
                  const matrix_expression<E2> &e2,
                  dense_proxy_tag,     row_major_tag,
                  M &m,
                  unknown_storage_tag, column_major_tag) {
        return axpy_prod2(e1, unknown_storage_tag(), row_major_tag(),
                          e2, dense_proxy_tag(),     column_major_tag(),
                          m,  unknown_storage_tag(), column_major_tag() );
      }

      template<class M, class E1, class E2>
      BOOST_UBLAS_INLINE
      M&
      axpy_prod2 (const matrix_expression<E1> &e1,
                  unknown_storage_tag, row_major_tag,
                  const matrix_expression<E2> &e2,
                  dense_proxy_tag,     row_major_tag,
                  M &m,
                  dense_proxy_tag,     column_major_tag) {
        return axpy_prod2(e1, unknown_storage_tag(), row_major_tag(),
                          e2, dense_proxy_tag(),     column_major_tag(),
                          m,  unknown_storage_tag(), column_major_tag() );
      }

      template<class M, class E1, class E2>
      BOOST_UBLAS_INLINE
      M&
      axpy_prod2 (const matrix_expression<E1> &e1,
                  unknown_storage_tag, row_major_tag,
                  const matrix_expression<E2> &e2,
                  unknown_storage_tag, row_major_tag,
                  M &m,
                  dense_proxy_tag,     column_major_tag) {
        return axpy_prod2(e1, unknown_storage_tag(), row_major_tag(),
                          e2, unknown_storage_tag(), row_major_tag(),
                          m,  dense_proxy_tag(),     row_major_tag() );
      }

      template<class M, class E1, class E2>
      BOOST_UBLAS_INLINE
      M&
      axpy_prod2 (const matrix_expression<E1> &e1,
                  unknown_storage_tag, row_major_tag,
                  const matrix_expression<E2> &e2,
                  unknown_storage_tag, row_major_tag,
                  M &m,
                  unknown_storage_tag, column_major_tag) {
        return axpy_prod2(e1, unknown_storage_tag(), column_major_tag(),
                          e2, unknown_storage_tag(), column_major_tag(),
                          m,  unknown_storage_tag(), column_major_tag() );
      }

      template<class M, class E1, class E2>
      BOOST_UBLAS_INLINE
      M&
      axpy_prod2 (const matrix_expression<E1> &e1,
                  unknown_storage_tag, row_major_tag,
                  const matrix_expression<E2> &e2,
                  unknown_storage_tag, column_major_tag,
                  M &m,
                  unknown_storage_tag, row_major_tag) {
        return axpy_prod2(e1, unknown_storage_tag(), row_major_tag(),
                          e2, unknown_storage_tag(), row_major_tag(),
                          m,  unknown_storage_tag(), row_major_tag());
      }

      template<class M, class E1, class E2>
      BOOST_UBLAS_INLINE
      M&
      axpy_prod2 (const matrix_expression<E1> &e1,
                  unknown_storage_tag, row_major_tag,
                  const matrix_expression<E2> &e2,
                  unknown_storage_tag, column_major_tag,
                  M &m,
                  unknown_storage_tag, column_major_tag) {
        return axpy_prod2(e1, unknown_storage_tag(), column_major_tag(),
                          e2, unknown_storage_tag(), column_major_tag(),
                          m,  unknown_storage_tag(), column_major_tag());
      }

      template<class M, class E1, class E2>
      BOOST_UBLAS_INLINE
      M&
      axpy_prod2 (const matrix_expression<E1> &e1,
                  unknown_storage_tag, column_major_tag,
                  const matrix_expression<E2> &e2,
                  unknown_storage_tag, row_major_tag,
                  M &m,
                  unknown_storage_tag, row_major_tag) {
        return axpy_prod2(e1, unknown_storage_tag(), row_major_tag(),
                          e2, unknown_storage_tag(), row_major_tag(),
                          m,  unknown_storage_tag(), row_major_tag());
      }

      template<class M, class E1, class E2>
      BOOST_UBLAS_INLINE
      M&
      axpy_prod2 (const matrix_expression<E1> &e1,
                  unknown_storage_tag, column_major_tag,
                  const matrix_expression<E2> &e2,
                  unknown_storage_tag, column_major_tag,
                  M &m,
                  dense_proxy_tag,     row_major_tag) {
        return axpy_prod2(e1, unknown_storage_tag(), column_major_tag(),
                          e2, unknown_storage_tag(), column_major_tag(),
                          m,  unknown_storage_tag(), column_major_tag());
      }

      template<class M, class E1, class E2>
      BOOST_UBLAS_INLINE
      M&
      axpy_prod2 (const matrix_expression<E1> &e1,
                  unknown_storage_tag, column_major_tag,
                  const matrix_expression<E2> &e2,
                  unknown_storage_tag, column_major_tag,
                  M &m,
                  unknown_storage_tag, row_major_tag) {
        return axpy_prod2(e1, unknown_storage_tag(), row_major_tag(),
                          e2, unknown_storage_tag(), row_major_tag(),
                          m,  unknown_storage_tag(), row_major_tag());
      }

      template<class M, class E1, class E2>
      BOOST_UBLAS_INLINE
      M&
      axpy_prod2 (const matrix_expression<E1> &e1,
                  unknown_storage_tag, column_major_tag,
                  const matrix_expression<E2> &e2,
                  unknown_storage_tag, row_major_tag,
                  M &m,
                  unknown_storage_tag, column_major_tag) {
        return axpy_prod2(e1, unknown_storage_tag(), column_major_tag(),
                          e2, unknown_storage_tag(), column_major_tag(),
                          m,  unknown_storage_tag(), column_major_tag());
      }

      // dispatcher
      template<class M, class E1, class E2>
      BOOST_UBLAS_INLINE
      M&
      axpy_prod2 (const matrix_expression<E1> &e1,
                  const matrix_expression<E2> &e2,
                  M &m, bool init = true) {
        typedef typename E1::storage_category     E1_storage_category;
        typedef typename E1::orientation_category E1_orientation_category;

        typedef typename E2::storage_category     E2_storage_category;
        typedef typename E2::orientation_category E2_orientation_category;

        typedef typename M::value_type value_type;
        typedef typename M::storage_category      M_storage_category;
        typedef typename M::orientation_category  M_orientation_category;

        if (init)
          m.assign (zero_matrix<value_type> (e1 ().size1 (), e2 ().size2 ()));
        return axpy_prod2 (e1, E1_storage_category (), E1_orientation_category (),
                           e2, E2_storage_category (), E2_orientation_category (),
                           m,  M_storage_category (),  M_orientation_category ());
      }

      template<typename T1, typename T2, typename MatrA, typename MatrB, typename MatrC>
      void gemm(const bool TransA, const bool TransB,
                const T1 &alpha, const MatrA &a, const MatrB &b,
                const T2 &beta, MatrC& c)
      {
        if (beta != 1.0 && beta != 0.0)
          c *= beta;
        if (alpha == 1.0) {
          if        (!TransA && !TransB) {
            axpy_prod2(a,        b,        c, beta==0.0);
          } else if (TransA  && !TransB) {
            axpy_prod2(trans(a), b,        c, beta==0.0);
          } else if (!TransA && TransB ) {
            axpy_prod2(a,        trans(b), c, beta==0.0);
          } else if (TransA  && TransB ) {
            axpy_prod2(trans(a), trans(b), c, beta==0.0);
          }
        } else {
          if        (!TransA && !TransB) {
            axpy_prod2(alpha * a,        b,        c, beta==0.0);
          } else if (TransA  && !TransB) {
            axpy_prod2(alpha * trans(a), b,        c, beta==0.0);
          } else if (!TransA && TransB ) {
            axpy_prod2(alpha * a,        trans(b), c, beta==0.0);
          } else if (TransA  && TransB ) {
            axpy_prod2(alpha * trans(a), trans(b), c, beta==0.0);
          }
        }
      }
    }
  }
}

// start of test structure

template<class mat1, class mat2>
double diff(const mat1& A, const mat2& B) {
  double result = 0;
  for (typename mat1::size_type i=0; i<A.size1 (); ++i) {
    for (typename mat1::size_type j=0; j<A.size2 (); ++j) {
      const double temp = A(i,j)-B(i,j);
      result += temp*temp;
    }
  }
  return sqrt(result);
}

// efficiently fill matrix depending on majority
template<class mat>
void fill_matrix(mat& A, ublas::column_major_tag) {
  int max_idx = static_cast<int>(std::min(A.size1(), A.size2()));
  for (int i=0; i<max_idx; ++i) {
    if (i-1500>=0) {
      A(i-1500, i) = 8;
    }
    if (i-1000>=0) {
      A(i-1000, i) = -5;
    }
    if (i-20>=0) {
      A(i-20, i) = -3;
    }
    if (i-1>=0) {
      A(i-1, i) = -2;
    }
    A(i, i) = 1;
    if (i+1<max_idx) {
      A(i+1, i) = -1;
    }
    if (i+20<max_idx) {
      A(i+20, i) = -2;
    }
    if (i+1000<max_idx) {
      A(i+1000, i) = -4;
    }
    if (i+1500<max_idx) {
      A(i+1500, i) = 7;
    }
  }
}

template<class mat>
void fill_matrix(mat& A, ublas::row_major_tag) {
  int max_idx = static_cast<int>(std::min(A.size1(), A.size2()));
  for (int i=0; i<max_idx; ++i) {
    if (i-1500>=0) {
      A(i-1500, i) = 7;
    }
    if (i-1000>=0) {
      A(i, i-1000) = -4;
    }
    if (i-20>=0) {
      A(i, i-20) = -2;
    }
    if (i-1>=0) {
      A(i, i-1) = -1;
    }
    A(i, i) = 1;
    if (i+1<max_idx) {
      A(i, i+1) = -2;
    }
    if (i+20<max_idx) {
      A(i, i+20) = -3;
    }
    if (i+1000<max_idx) {
      A(i, i+1000) = -5;
    }
    if (i+1500<max_idx) {
      A(i+1500, i) = 8;
    }
  }
}

template<class mat1, class mat2, class mat3>
BOOST_UBLAS_TEST_DEF ( test_axpy_prod )
{
  mat1 A(n, m);
  A.clear();
  fill_matrix(A, typename mat1::orientation_category());

  mat2 B(m, o);
  B.clear();
  fill_matrix(B, typename mat2::orientation_category());

  if (n<=3) {
    std::cerr << A <<"\n";
    std::cerr << B <<"\n";
  }
  clock_t axpy_prod2_t, sparse_prod_t, axpy_prod_t, prod_t;
  {
    mat3 C(n, o);
    axpy_prod2(A, B, C);
  }

  mat3 CC(n, o);
  std::cerr << ".";
  {
    clock_t start = clock();
    mat3 C1(n, o);
    axpy_prod2(A, B, C1);
    axpy_prod2_t = clock()-start;
    CC = C1;
    if (n<=3)
      std::cerr << C1;
  }
  std::cerr.width(10);
  std::cerr  << axpy_prod2_t << ".";
  {
    clock_t start = clock();
    mat3 C2(n, o);
    sparse_prod(A, B, C2);
    sparse_prod_t = clock()-start;
    if (n<=3) {
      std::cerr << C2;
      assert(diff(CC, CC)<TOL);
    }
  }
  std::cerr.width(10);
  std::cerr << sparse_prod_t << ".";
  {
    mat3 C(n, o);
    clock_t start = clock();
    axpy_prod(A, B, C);
    axpy_prod_t = clock()-start;
  }
  std::cerr.width(10);
  std::cerr << axpy_prod_t << ".";

  {
    mat3 C(n, o);
    clock_t start = clock();
    C = prod(A, B);
    prod_t = clock()-start;;
  }
  gemm(true, true, 1.0, A, B, 0.0, CC);
  std::cerr.width(10);
  std::cerr << prod_t << ".";

  std::cerr << " spdup: " << std::fixed << std::setprecision(2) << std::setw(7) << double(sparse_prod_t)/double(axpy_prod2_t);
  std::cerr << " "       << std::setprecision(2) << std::setw(7) << double(axpy_prod_t)  /double(axpy_prod2_t);
  std::cerr << " "       << std::setprecision(2) << std::setw(7) << double(prod_t)       /double(axpy_prod2_t);
  std::cerr << "\n";

}

int main() {

  // typedefs are needed as macros do not work with "," in template arguments
  typedef ublas::compressed_matrix<double, ublas::row_major>     commat_doub_rowmaj;
  typedef ublas::compressed_matrix<double, ublas::column_major>  commat_doub_colmaj;
  typedef ublas::matrix<double, ublas::row_major>                denmat_doub_rowmaj;
  typedef ublas::matrix<double, ublas::column_major>             denmat_doub_colmaj;
  typedef ublas::mapped_matrix<double, ublas::row_major>         mapmat_doub_rowmaj;
  typedef ublas::mapped_matrix<double, ublas::column_major>      mapmat_doub_colmaj;
  typedef ublas::coordinate_matrix<double, ublas::row_major>     cormat_doub_rowmaj;
  typedef ublas::coordinate_matrix<double, ublas::column_major>  cormat_doub_colmaj;
  typedef ublas::mapped_vector_of_mapped_vector<double, ublas::row_major> mvmv_doub_rowmaj;
  typedef ublas::mapped_vector_of_mapped_vector<double, ublas::column_major> mvmv_doub_colmaj;


  BOOST_UBLAS_TEST_BEGIN();

  std::cerr<<"commat_doub_rowmaj\n";
  test_axpy_prod<commat_doub_rowmaj, commat_doub_rowmaj, commat_doub_rowmaj>(test_fails_);
  test_axpy_prod<commat_doub_rowmaj, commat_doub_rowmaj, denmat_doub_rowmaj>(test_fails_);
  test_axpy_prod<commat_doub_rowmaj, commat_doub_rowmaj, commat_doub_colmaj>(test_fails_);
  test_axpy_prod<commat_doub_rowmaj, commat_doub_rowmaj, denmat_doub_colmaj>(test_fails_);

  test_axpy_prod<commat_doub_rowmaj, denmat_doub_rowmaj, commat_doub_rowmaj>(test_fails_);
  test_axpy_prod<commat_doub_rowmaj, denmat_doub_rowmaj, denmat_doub_rowmaj>(test_fails_);
  test_axpy_prod<commat_doub_rowmaj, denmat_doub_rowmaj, commat_doub_colmaj>(test_fails_);
  test_axpy_prod<commat_doub_rowmaj, denmat_doub_rowmaj, denmat_doub_colmaj>(test_fails_);

  test_axpy_prod<commat_doub_rowmaj, commat_doub_colmaj, commat_doub_rowmaj>(test_fails_);
  test_axpy_prod<commat_doub_rowmaj, commat_doub_colmaj, denmat_doub_rowmaj>(test_fails_);
  test_axpy_prod<commat_doub_rowmaj, commat_doub_colmaj, commat_doub_colmaj>(test_fails_);
  test_axpy_prod<commat_doub_rowmaj, commat_doub_colmaj, denmat_doub_colmaj>(test_fails_);

  test_axpy_prod<commat_doub_rowmaj, denmat_doub_colmaj, commat_doub_rowmaj>(test_fails_);
  test_axpy_prod<commat_doub_rowmaj, denmat_doub_colmaj, denmat_doub_rowmaj>(test_fails_);
  test_axpy_prod<commat_doub_rowmaj, denmat_doub_colmaj, commat_doub_colmaj>(test_fails_);
  test_axpy_prod<commat_doub_rowmaj, denmat_doub_colmaj, denmat_doub_colmaj>(test_fails_);

  std::cerr<<"commat_doub_colmaj\n";
  test_axpy_prod<commat_doub_colmaj, commat_doub_rowmaj, commat_doub_rowmaj>(test_fails_);
  test_axpy_prod<commat_doub_colmaj, commat_doub_rowmaj, denmat_doub_rowmaj>(test_fails_);
  test_axpy_prod<commat_doub_colmaj, commat_doub_rowmaj, commat_doub_colmaj>(test_fails_);
  test_axpy_prod<commat_doub_colmaj, commat_doub_rowmaj, denmat_doub_colmaj>(test_fails_);

  test_axpy_prod<commat_doub_colmaj, denmat_doub_rowmaj, commat_doub_rowmaj>(test_fails_);
  test_axpy_prod<commat_doub_colmaj, denmat_doub_rowmaj, denmat_doub_rowmaj>(test_fails_);
  test_axpy_prod<commat_doub_colmaj, denmat_doub_rowmaj, commat_doub_colmaj>(test_fails_);
  test_axpy_prod<commat_doub_colmaj, denmat_doub_rowmaj, denmat_doub_colmaj>(test_fails_);

  test_axpy_prod<commat_doub_colmaj, commat_doub_colmaj, commat_doub_rowmaj>(test_fails_);
  test_axpy_prod<commat_doub_colmaj, commat_doub_colmaj, denmat_doub_rowmaj>(test_fails_);
  test_axpy_prod<commat_doub_colmaj, commat_doub_colmaj, commat_doub_colmaj>(test_fails_);
  test_axpy_prod<commat_doub_colmaj, commat_doub_colmaj, denmat_doub_colmaj>(test_fails_);

  test_axpy_prod<commat_doub_colmaj, denmat_doub_colmaj, commat_doub_rowmaj>(test_fails_);
  test_axpy_prod<commat_doub_colmaj, denmat_doub_colmaj, denmat_doub_rowmaj>(test_fails_);
  test_axpy_prod<commat_doub_colmaj, denmat_doub_colmaj, commat_doub_colmaj>(test_fails_);
  test_axpy_prod<commat_doub_colmaj, denmat_doub_colmaj, denmat_doub_colmaj>(test_fails_);

  std::cerr<<"denmat_doub_rowmaj\n";
  test_axpy_prod<denmat_doub_rowmaj, commat_doub_rowmaj, commat_doub_rowmaj>(test_fails_);
  test_axpy_prod<denmat_doub_rowmaj, commat_doub_rowmaj, denmat_doub_rowmaj>(test_fails_);
  test_axpy_prod<denmat_doub_rowmaj, commat_doub_rowmaj, commat_doub_colmaj>(test_fails_);
  test_axpy_prod<denmat_doub_rowmaj, commat_doub_rowmaj, denmat_doub_colmaj>(test_fails_);

  test_axpy_prod<denmat_doub_rowmaj, denmat_doub_rowmaj, commat_doub_rowmaj>(test_fails_);
  test_axpy_prod<denmat_doub_rowmaj, denmat_doub_rowmaj, denmat_doub_rowmaj>(test_fails_);
  test_axpy_prod<denmat_doub_rowmaj, denmat_doub_rowmaj, commat_doub_colmaj>(test_fails_);
  test_axpy_prod<denmat_doub_rowmaj, denmat_doub_rowmaj, denmat_doub_colmaj>(test_fails_);

  test_axpy_prod<denmat_doub_rowmaj, commat_doub_colmaj, commat_doub_rowmaj>(test_fails_);
  test_axpy_prod<denmat_doub_rowmaj, commat_doub_colmaj, denmat_doub_rowmaj>(test_fails_);
  test_axpy_prod<denmat_doub_rowmaj, commat_doub_colmaj, commat_doub_colmaj>(test_fails_);
  test_axpy_prod<denmat_doub_rowmaj, commat_doub_colmaj, denmat_doub_colmaj>(test_fails_);

  test_axpy_prod<denmat_doub_rowmaj, denmat_doub_colmaj, commat_doub_rowmaj>(test_fails_);
  test_axpy_prod<denmat_doub_rowmaj, denmat_doub_colmaj, denmat_doub_rowmaj>(test_fails_);
  test_axpy_prod<denmat_doub_rowmaj, denmat_doub_colmaj, commat_doub_colmaj>(test_fails_);
  test_axpy_prod<denmat_doub_rowmaj, denmat_doub_colmaj, denmat_doub_colmaj>(test_fails_);

  std::cerr<<"denmat_doub_colmaj\n";
  test_axpy_prod<denmat_doub_colmaj, commat_doub_rowmaj, commat_doub_rowmaj>(test_fails_);
  test_axpy_prod<denmat_doub_colmaj, commat_doub_rowmaj, denmat_doub_rowmaj>(test_fails_);
  test_axpy_prod<denmat_doub_colmaj, commat_doub_rowmaj, commat_doub_colmaj>(test_fails_);
  test_axpy_prod<denmat_doub_colmaj, commat_doub_rowmaj, denmat_doub_colmaj>(test_fails_);

  test_axpy_prod<denmat_doub_colmaj, denmat_doub_rowmaj, commat_doub_rowmaj>(test_fails_);
  test_axpy_prod<denmat_doub_colmaj, denmat_doub_rowmaj, denmat_doub_rowmaj>(test_fails_);
  test_axpy_prod<denmat_doub_colmaj, denmat_doub_rowmaj, commat_doub_colmaj>(test_fails_);
  test_axpy_prod<denmat_doub_colmaj, denmat_doub_rowmaj, denmat_doub_colmaj>(test_fails_);

  test_axpy_prod<denmat_doub_colmaj, commat_doub_colmaj, commat_doub_rowmaj>(test_fails_);
  test_axpy_prod<denmat_doub_colmaj, commat_doub_colmaj, denmat_doub_rowmaj>(test_fails_);
  test_axpy_prod<denmat_doub_colmaj, commat_doub_colmaj, commat_doub_colmaj>(test_fails_);
  test_axpy_prod<denmat_doub_colmaj, commat_doub_colmaj, denmat_doub_colmaj>(test_fails_);

  test_axpy_prod<denmat_doub_colmaj, denmat_doub_colmaj, commat_doub_rowmaj>(test_fails_);
  test_axpy_prod<denmat_doub_colmaj, denmat_doub_colmaj, denmat_doub_rowmaj>(test_fails_);
  test_axpy_prod<denmat_doub_colmaj, denmat_doub_colmaj, commat_doub_colmaj>(test_fails_);
  test_axpy_prod<denmat_doub_colmaj, denmat_doub_colmaj, denmat_doub_colmaj>(test_fails_);

//   std::cerr<<"other stuff\n";
//   test_axpy_prod<mapmat_doub_rowmaj, mapmat_doub_colmaj, denmat_doub_rowmaj>(test_fails_);
//   test_axpy_prod<mapmat_doub_rowmaj, mapmat_doub_colmaj, denmat_doub_colmaj>(test_fails_);
//   test_axpy_prod<mapmat_doub_rowmaj, mapmat_doub_colmaj, mapmat_doub_rowmaj>(test_fails_);
//   test_axpy_prod<mapmat_doub_rowmaj, mapmat_doub_colmaj, mapmat_doub_colmaj>(test_fails_);

//   test_axpy_prod<mapmat_doub_rowmaj, mapmat_doub_rowmaj, denmat_doub_rowmaj>(test_fails_);
//   test_axpy_prod<mapmat_doub_rowmaj, mapmat_doub_rowmaj, denmat_doub_colmaj>(test_fails_);
//   test_axpy_prod<mapmat_doub_rowmaj, mapmat_doub_rowmaj, mapmat_doub_rowmaj>(test_fails_);
//   test_axpy_prod<mapmat_doub_rowmaj, mapmat_doub_rowmaj, mapmat_doub_colmaj>(test_fails_);

//   test_axpy_prod<mapmat_doub_colmaj, mapmat_doub_colmaj, mapmat_doub_rowmaj>(test_fails_);
//   test_axpy_prod<mapmat_doub_colmaj, mapmat_doub_colmaj, mapmat_doub_colmaj>(test_fails_);
//   test_axpy_prod<mapmat_doub_colmaj, mapmat_doub_colmaj, denmat_doub_rowmaj>(test_fails_);
//   test_axpy_prod<mapmat_doub_colmaj, mapmat_doub_colmaj, denmat_doub_colmaj>(test_fails_);

//   test_axpy_prod<mapmat_doub_colmaj, mapmat_doub_rowmaj, mapmat_doub_rowmaj>(test_fails_);
//   test_axpy_prod<mapmat_doub_colmaj, mapmat_doub_rowmaj, mapmat_doub_colmaj>(test_fails_);
//   test_axpy_prod<mapmat_doub_colmaj, mapmat_doub_rowmaj, denmat_doub_rowmaj>(test_fails_);
//   test_axpy_prod<mapmat_doub_colmaj, mapmat_doub_rowmaj, denmat_doub_colmaj>(test_fails_);

//   test_axpy_prod<cormat_doub_rowmaj, cormat_doub_colmaj, denmat_doub_rowmaj>(test_fails_);
//   test_axpy_prod<cormat_doub_rowmaj, cormat_doub_colmaj, denmat_doub_colmaj>(test_fails_);
//   test_axpy_prod<cormat_doub_rowmaj, cormat_doub_colmaj, cormat_doub_rowmaj>(test_fails_);
//   test_axpy_prod<cormat_doub_rowmaj, cormat_doub_colmaj, cormat_doub_colmaj>(test_fails_);

//   test_axpy_prod<cormat_doub_rowmaj, cormat_doub_rowmaj, denmat_doub_rowmaj>(test_fails_);
//   test_axpy_prod<cormat_doub_rowmaj, cormat_doub_rowmaj, denmat_doub_colmaj>(test_fails_);
//   test_axpy_prod<cormat_doub_rowmaj, cormat_doub_rowmaj, cormat_doub_rowmaj>(test_fails_);
//   test_axpy_prod<cormat_doub_rowmaj, cormat_doub_rowmaj, cormat_doub_colmaj>(test_fails_);

//   test_axpy_prod<cormat_doub_colmaj, cormat_doub_colmaj, cormat_doub_rowmaj>(test_fails_);
//   test_axpy_prod<cormat_doub_colmaj, cormat_doub_colmaj, cormat_doub_colmaj>(test_fails_);
//   test_axpy_prod<cormat_doub_colmaj, cormat_doub_colmaj, denmat_doub_rowmaj>(test_fails_);
//   test_axpy_prod<cormat_doub_colmaj, cormat_doub_colmaj, denmat_doub_colmaj>(test_fails_);

//   test_axpy_prod<cormat_doub_colmaj, cormat_doub_rowmaj, cormat_doub_rowmaj>(test_fails_);
//   test_axpy_prod<cormat_doub_colmaj, cormat_doub_rowmaj, cormat_doub_colmaj>(test_fails_);
//   test_axpy_prod<cormat_doub_colmaj, cormat_doub_rowmaj, denmat_doub_rowmaj>(test_fails_);
//   test_axpy_prod<cormat_doub_colmaj, cormat_doub_rowmaj, denmat_doub_colmaj>(test_fails_);

  BOOST_UBLAS_TEST_END();
}
